/*! \file fourier.cpp
 *  \brief Implementation of Fourier-domain functions for the symmetric group
 */

#include "fourier.h"
#include "util.h"

#include <vector>
#include <cmath>
#include <iostream>

#include <itpp/itbase.h>

#define SQ(x) ((x)*(x))

using namespace std;

using namespace itpp;

#define Matrix mat

/*! \brief Find the nonzero elements in row (or column) j of the irrep
 *   at (i, i+1) in the gz basis given by the specified tableaux
 * 
 *  This function exists to allow efficient "sparse" computation with these
 *   irreps
 *  \todo Currently this is Theta(irrep dimension), which does not take
 *   advantage of sparseness. Reimplement with map for Theta(lg d) or
 *   with hash_map for Theta(1)
 */
static vector<pair<int, double> >
gz_irrep_adj_swap_eles(vector<Tableau> tabs, int i, int j)
{
        vector<pair<int, double> > elements;

        elements.push_back(make_pair(j, 1. / tabs[j].axial_dist(i, i + 1)));
        
        pair<int, int> i_pos = tabs[j].pos(i);
        pair<int, int> i1_pos = tabs[j].pos(i + 1);
        if (i_pos.first != i1_pos.first
            && i_pos.second != i1_pos.second) {
                Tableau ts = tabs[j].swap(i);
                unsigned int k;
                for (k = 0; k < tabs.size(); k++)
                        if (tabs[k] == ts)
                                break;
                elements.push_back(make_pair(k, 
                       sqrt(1. - 1./SQ(tabs[j].axial_dist(i, i+1)))));
        }

        return elements;
}

Matrix gz_irrep(Partition part, Permutation p)
{
        return mult_irrep_left(part, p, eye(Tableau::count(part)));
}

/// \todo right multiplication in wrong order
static Matrix mult_irrep(Partition part, Permutation p, Matrix m, bool left_mult)
{
        vector<Tableau> tabs = Tableau::list(part);
        /// \todo test this for identity
        vector<Permutation> factors = p.factor_adj_swaps();
        Matrix res;

        if (factors.size() == 0)
                return m;

        res = m;

        int d = res.rows();
        for (int i = factors.size() - 1; i >= 0; i--) {
                Matrix prod (d, d);
                prod.zeros();
                // multiply by irrep at this swap
                for (int j = 0; j < d; j++) {
                        vector<pair<int, double> > eles
                                 = gz_irrep_adj_swap_eles(tabs,
                                                factors[i].min_cycle_ele(), j);
                        for (int k = 0; k < d; k++) {
                                double sum = 0;
                          
                                for (unsigned int m = 0; m < eles.size(); m++) {
                                        if (left_mult)
                                          sum += eles[m].second
                                               *res(eles[m].first, k);
                                        else
                                          sum += eles[m].second
                                               *res(k, eles[m].first);
                                }
                                if (left_mult)
                                        prod.set(j, k, sum);
                                else
                                        prod.set(k, j, sum);
                        }
                }
                res = prod;
        }
        
        return res;
}

Matrix mult_irrep_left(Partition part, Permutation p, Matrix m)
{
        return mult_irrep(part, p, m, true);
}

Matrix mult_irrep_right(Partition part, Permutation p, Matrix m)
{
        return mult_irrep(part, p, m, false);
}


vector<int> character(Partition part)
{
        vector<int> ch;

        for (unsigned int i = 0; i < fact(part.num()); i++) {
                Permutation p = Permutation::from_index(part.num(), i);
                ch.push_back(RND(trace(gz_irrep(part, p))));
        }

        return ch;
}

/// \todo Check ch0.size() == ch1.size()
int char_prod(vector<int> ch0, vector<int> ch1)
{
        int sum = 0;

        for (unsigned int i = 0; i < ch0.size(); i++)
                sum += ch0[i] * ch1[i];

        return sum/ch0.size();
}


Matrix marginal_rep(Partition part, Permutation p)
{
        vector<Permutation> tabloids = tabloid_list(part);

        Matrix m (tabloids.size(), tabloids.size());
        m.zeros();

        for (unsigned int i = 0; i < tabloids.size(); i++)
                for (unsigned int j = 0; j < tabloids.size(); j++)
                        if (tabloid_equal(part, p.of(tabloids[i]), tabloids[j]))
                                m.set(j, i, 1);

        return m;
}

Matrix direct_sum(Matrix m0, Matrix m1)
{
        Matrix z0 (m0.rows(), m1.cols());
        Matrix z1 (m1.rows(), m0.cols());

        z0.zeros();
        z1.zeros();

        return concat_vertical(concat_horizontal(m0, z0),
                               concat_horizontal(z1, m1));
}

/// \todo is there a built-in way to do this?
Matrix direct_sum(vector<Matrix> blocks)
{
        if (blocks.size() == 0)
                /// \todo throw error
                cout << "Empty direct sum" << endl;

        Matrix m = blocks[0];

        for (unsigned int i = 1; i < blocks.size(); i++)
                m = direct_sum(m, blocks[i]);

        return m;
}

Matrix multi_direct_sum(Matrix m, int n)
{
        Matrix s = m;

        for (int i = 1; i < n; i++)
                s = direct_sum(s, m);

        return s;
}

Matrix kron_factor_id(Matrix p, int d)
{
        int pd = p.rows();
        Matrix m (pd/d, pd/d);

        for (int i = 0; i < pd/d; i++)
                for (int j = 0; j < pd/d; j++)
                        m.set(i, j, p(i*d, j*d));

        return m;
}

Matrix normalize_rows(Matrix m)
{
        vec norms = sqrt(sum_sqr(m, 2));
        
        return elem_div(m, repmat(norms, 1, m.cols(), false));
}

/*! \brief Compute part of an orthogonal intertwining operator between two
 *   orthogonal representations
 *
 *  X0 and X1 should be elements of a representation that correspond to two
 *  different elements that generate S_n (such as a transposition and a
 *  complete cycle). Y0 and Y1 should be elements of a different representation
 *  that correspond to the same elements of S_n.
 *  z should be the multiplicity of the X representation in the Y
 *  representation (the dimension of the Y representation should be at least
 *  the dimension of the X representation).
 *
 *  In typical use, X is an irrep, Y is a marginal rep, and z is a Koskta
 *  number.
 *
 *  The matrix returned has orthonormal rows and forms an intertwining
 *  operator when stacked with other such matrices generated from
 *  X representations that from a decomposition of the representation Y.
 */ 
static Matrix intertwine(Matrix X0, Matrix X1, Matrix Y0, Matrix Y1, int z)
{
        /// \todo check size consistency
        int d = X0.rows();
        int dnu = Y0.rows();

        Matrix K0 = kron(eye(d), multi_direct_sum(Y0, z))
                  - kron(transpose(X0), eye(z*dnu));
        Matrix K1 = kron(eye(d), multi_direct_sum(Y1, z))
                  - kron(transpose(X1), eye(z*dnu));
        Matrix K = concat_vertical(K0, K1);


        // find z**2 dimensional nullspace of K
        Matrix U, V;
        vec s;
        svd(K, U, s, V);

        // choose a random element of that nullspace
        vec rand_vec = randu(z*z);
        // true means rand_vec is a row vector
        Matrix rand_mat = repmat(rand_vec, z*dnu*d, 1, true);
        Matrix null_basis = V.get_cols(V.cols() - z*z, V.cols() - 1);
        // a random linear combination of null space basis vectors
        vec v = sum(elem_mult(null_basis, rand_mat), 2);
        Matrix R = reshape(v, z*dnu, d);

        // orthogonalize
        Matrix R1 = R * transpose(R);

        Matrix M = kron_factor_id(R1, dnu);

        vec evals;
        Matrix S;
        eig_sym(M, evals, S);

        S = kron(S, eye(dnu));

        return normalize_rows(transpose(S) * R);
}

/*! \brief Vertically concatenate given vector of matrices */
static Matrix concat_vertical(vector<Matrix> ms)
{
        Matrix m = ms[0];

        for (unsigned int i = 1; i < ms.size(); i++)
                m = concat_vertical(m, ms[i]);

        return m;
}

Matrix marg_intertwine(Partition part, vector<Partition> dom_parts)
{
        vector<Matrix> subintertwiners;
        Permutation p0 = Permutation::swap(part.num(), 0, 1);
        Permutation p1 = Permutation::cont_cycle(part.num(), 0, part.num());
        Matrix m_rep0 = marginal_rep(part, p0);
        Matrix m_rep1 = marginal_rep(part, p1);

        for (unsigned int i = 0; i < dom_parts.size(); i++) {
                // is this check a speed up?
                if (!dom_parts[i].dominates(part))
                        continue;

                int k = kostka(part, dom_parts[i]);
                Matrix irrep0 = gz_irrep(dom_parts[i], p0);
                Matrix irrep1 = gz_irrep(dom_parts[i], p1);
                subintertwiners
                    .push_back(intertwine(m_rep0, m_rep1, irrep0, irrep1, k));
        }

        return concat_vertical(subintertwiners);
}
