/*! \file fourier.h
 *  \brief Declaration of Fourier-domain functions for the symmetric group
 */

#ifndef FOURIER_H
#define FOURIER_H

//#include "matrix.h"
#include "partition.h"

#include <vector>
#include <itpp/itbase.h>

#define Matrix itpp::mat

/*! \brief Construct the irreducible representation matrix in the
 *  Gel'fand-Tsetlin basis at the given permutation
 */
Matrix gz_irrep(Partition part, Permutation p);

/// Multiply a matrix by the partition part irrep at the permutation p
/*! This function is Theta(nd), where d is the dimension of the matrix and
 *  n is the number of adjacent transposition factors of the irrep
 *  \todo make sure above is true
 */
Matrix mult_irrep_left(Partition part, Permutation p, Matrix m);

/// Like mult_irrep_left, but perform right multiplication instead
Matrix mult_irrep_right(Partition part, Permutation p, Matrix m);

/// The character of the irrep at partition part
std::vector<int> character(Partition part);

/// The inner product of characters of two irreps
int char_prod(std::vector<int> ch0, std::vector<int> ch1);

/// Compute the direct sum a given vector of matrices */
Matrix direct_sum(std::vector<Matrix> ms);

/// Compute the multiple direct sum m (+) ... (+) m (n times)
Matrix multi_direct_sum(Matrix m, int n);

/*! \brief Compute an orthogonal intertwining operator between the matrix of
 *  marginals at partition part and the matrices of GZ basis irreps at
 *  the partitions dom_parts
 *
 *  The partitions from dom_parts are used with multiplicities given by the
 *  Kostka numbers; any that do not dominate part will not be used.
 *  Commonly, dom_parts will be all partitions above (and including part)
 *  in the dominance ordering.
 */
Matrix marg_intertwine(Partition part, std::vector<Partition> dom_parts);

/*! \brief Compute the the marginal representation of S_n at the permutation
 *  p for the partition part
 *
 *  The (i,j) entry of the resulting matrix is 1 if p(tabloid_j) = tabloid_i,
 *  where tabloids are indexed in the order given by tabloid_list, composition
 *  means applying the map p to each element of the tabloid, and equality
 *  means tabloid equivalance. Note the "swiched" order of indices, which is
 *  necessary for the composition order of matrices to follow that of
 *  permutations.
 */
Matrix marginal_rep(Partition part, Permutation p);

#undef Matrix

#endif
