/*! \file func.cpp
 *  \brief Implementation of explicit real-valued functions on permutations
 */

#include <sstream>
#include <cstring>
#include <iostream>
#include <assert.h>

#include "func.h"
#include "permutation.h"
#include "partition.h"
#include "fourier.h"
#include "util.h"
#include "fourierfunc.h"


using namespace std;

#define Matrix itpp::mat

Func::Func(int n)
{
        nelements = n;
        funclen = (int)fact(n);
        func = new val_t[funclen];
}

Func &Func::operator = (const Func &f)
{
        nelements = f.nelements;
        funclen = f.funclen;
        delete [] func;
        func = new val_t[funclen];
        memcpy(func, f.func, sizeof(val_t)*funclen);

        return *this;
}

int Func::num_elements() const
{
        return nelements;
}

Func &Func::operator += (const Func &f)
{
        for (int i = 0; i < funclen; i++)
                func[i] += f.func[i];

        return *this;
}

Func::Func(const Func &f)
 : func (NULL)
{
        *this = f;
}

Func::~Func()
{
        delete [] func;
}

string Func::str()
{
        ostringstream out;
        int max = MAX_PRINT_ELEMENTS;
        int i;

        if (funclen < max)
                max = funclen;

        for (i = 0; i < max; i++) {
                out << Permutation::from_index(nelements, i).str() << " "
                    << func[i];
                if (i != max - 1)
                        out << endl;
        }

        if (i < funclen)
                out << endl << "...";
                
        return out.str();
}

Func Func::zero(int n)
{
        Func f(n);

        for (int i = 0; i < f.funclen; i++)
                f.func[i] = 0;

        return f;
}


Func::val_t Func::operator () (const Permutation &sigma) const
{
        return func[sigma.index()];
}

Func *Func::operator * (const Func &other)
{
        /// \todo check dimensions
        Func *f = new Func(nelements);

        for (int i = 0; i < funclen; i++)
                f->func[i] = this->func[i] * other.func[i];

        return f;
}

Func *Func::convolve(const Func &other)
{
        /// \todo check dimensions
        /// \todo need a real test of this
        Func *f = new Func(nelements);

        for (int i = 0; i < funclen; i++) {
                Permutation s_i = Permutation::from_index(nelements, i);
                val_t sum = 0;
                for (int j = 0; j < funclen; j++) {
                        Permutation s_j = Permutation::from_index(nelements, j);
                        sum += (*this)(s_i.of(s_j.inverse())) * other(s_j);
                }
                f->func[i] = sum;
        }

        return f;
}

Func *Func::shift(const Permutation &p0, const Permutation &p1)
{
        Func *f = new Func(nelements);

        for (int i = 0; i < funclen; i++) {
                Permutation p = Permutation::from_index(nelements, i);
                f->func[i] = func[p0.inverse().of(p.of(p1)).index()];
        }
        
        return f;
}

FourierFunc Func::fourier() const
{
        FourierFunc ft = FourierFunc::zero(nelements);


        vector<Partition> parts = ft.ind_parts();
        for (unsigned int i = 0; i < parts.size(); i++) {
                Matrix fc (Tableau::count(parts[i]),
                           Tableau::count(parts[i]));
                fc.zeros();

                for (int j = 0; j < funclen; j++) {
                        Permutation p = Permutation::from_index(nelements, j);
                        fc += gz_irrep(parts[i], p) * func[j];
                }

                ft.set(i, fc);
        }

        return ft;
}

void Func::set(const Permutation &p, val_t v)
{
        func[p.index()] = v;
}

/// \todo more general restriction?
Func Func::restrict_to(int i) const
{
        Func f(nelements - 1);

        for (int j = 0; j < f.funclen; j++) {
                // note that this is a permutation in S_n given by an index
                //  of a permutation in S_(n-1) (and relying on the order
                //  of permutations)
                Permutation p = Permutation::from_index(nelements, j);
                Permutation cc = Permutation::cont_cycle(nelements, i,
                                                         nelements);
                f.func[j] = (*this)(cc.of(p));
        }

        return f;
}

Func Func::embed(int i) const
{
        Func f = zero(nelements + 1);

        assert(i < nelements);

        Permutation cc = Permutation::cont_cycle(nelements + 1, i,
                                                 nelements + 1);
        for (int j = 0; j < funclen; j++) {
                // here we rely on the index of a permutation in S_n being
                // the correct index in S_(n+1)
                Permutation p = Permutation::from_index(nelements + 1, j);
                f.func[cc.of(p).index()] = func[j];
        }

        return f;
}

/*! \todo disallow 0-element FourierFunc construction */
FourierFunc Func::fft() const
{
        FourierFunc ft = FourierFunc::zero(nelements);
        vector<FourierFunc> subfts;

        // the trivial case
        if (nelements == 1) {
                Matrix m (1, 1);
                m.set(0, 0, func[0]);
                ft.set(0, m);

                return ft;
        }

        // look at n ways of restricting this func to S_(n-1) and compute
        //  the fft
        for (int i = 0; i < nelements; i++)
                subfts.push_back(restrict_to(i).fft());

        vector<Partition> parts = ft.ind_parts();
        for (unsigned int i = 0; i < parts.size(); i++) {
                Matrix fc (Tableau::count(parts[i]),
                           Tableau::count(parts[i]));
                fc.zeros();
                vector<Partition> subparts = parts[i].minus();

                for (int j = 0; j < nelements; j++) {
                        // is this the right c.c.?
                        Permutation cc = Permutation::cont_cycle(nelements,
                                                              j, nelements);
                        vector<Matrix> subfcs;
                        for (unsigned int k = 0; k < subparts.size(); k++)
                                subfcs.push_back(subfts[j].at(subparts[k]));
                        Matrix subfc = direct_sum(subfcs);
                        fc += mult_irrep_left(parts[i], cc, subfc);
                }
                ft.set(i, fc);
        }

        return ft;
}

Func::val_t Func::marginalize(Partition part, Permutation p0, Permutation p1)
{
        val_t marginal = 0;

        for (int i = 0; i < funclen; i++) {
                Permutation p = Permutation::from_index(nelements, i);

                if (tabloid_equal(part, p.of(p0), p1))
                        marginal += func[i];
        }

        return marginal;
}

Matrix Func::marginals(Partition part)
{
        vector<Permutation> tabloids = tabloid_list(part);

        Matrix m (tabloids.size(), tabloids.size());
        m.zeros();
        for (unsigned int i = 0; i < tabloids.size(); i++)
                for (unsigned int j = 0; j < tabloids.size(); j++)
                        m.set(j, i, marginalize(part, tabloids[i],
                                                tabloids[j]));

        return m;
}

Func::val_t Func::ordered(int i, int j)
{
        val_t sum = 0;
        for (int k = 0; k < funclen; k++) {
                Permutation p = Permutation::from_index(nelements, k);
                if (p(i) < p(j))
                        sum += func[k];
        }

        return sum;
}

Matrix Func::orderings()
{
        Matrix m (nelements, nelements);
        m.zeros();

        for (int i = 0; i < nelements; i++)
                for (int j = 0; j < nelements; j++)
                        m.set(i, j, ordered(i, j));

        return m;
}

Func *Func::join(Func *d0, Func *d1)
{
        Func *d = new Func(d0->nelements + d1->nelements);

        // initialize to zeros
        for (int i = 0; i < d->funclen; i++)
                d->func[i] = 0;

        // fill in first-order independent values
        for (int i = 0; i < d0->funclen; i++) {
                for (int j = 0; j < d1->funclen; j++) {
                        Permutation p = Permutation::from_index(d0->nelements, i)
                                .direct_sum(Permutation::from_index(d1->nelements, j));

                        d->func[p.index()] = d0->func[i] * d1->func[j];
                }
        }

        return d;
}

pair<Func *, Func*> Func::split(int n/*, int *is*/)
{
        Func *d0 = new Func(n);
        Func *d1 = new Func(nelements - n);
        int parts0[1 + n];
        int parts1[1 + nelements - n];

        parts0[0] = nelements - n;
        for (int i = 1; i < 1 + n; i++)
                parts0[i] = 1;
        parts1[0] = n;
        for (int i = 1; i < 1 + nelements - n; i++)
                parts1[i] = 1;

        Partition part0 (1 + n, parts0);
        Partition part1 (1 + nelements - n, parts1);

        for (int i = 0; i < d0->funclen; i++) {
                Permutation p = Permutation::from_index(n, i);
                d0->func[i] = Func::marginalize(part0, Permutation::identity(nelements),
                                                 Permutation::identity(nelements - n)
                                                  .direct_sum(p));
        }
        for (int i = 0; i < d1->funclen; i++) {
                Permutation p = Permutation::from_index(n, i);
                d1->func[i] = Func::marginalize(part1, Permutation::identity(nelements),
                                                 Permutation::identity(n).direct_sum(p));
        }

        return make_pair(d0, d1);
}



void Func::normalize()
{
        val_t sum = 0;
        
        for (int i = 0; i < funclen; i++)
                sum += func[i];

        for (int i = 0; i < funclen; i++)
                func[i] /= sum;
}
