/*! \file func.h
 *  \brief Definition of class for real-valued functions on permutations
 */

#ifndef FUNC_H
#define FUNC_H

#include <string>

#include "partition.h"
#include "permutation.h"
//#include "matrix.h"
#include <itpp/itbase.h>

#define Matrix itpp::mat

class FourierFunc;

/// Real-valued explicitly-stored functions on permutations
/*! \todo add min, max */
class Func {


public:
        // The type that this function maps to
        typedef double val_t;

        Func(const Func &f);

        Func &operator = (const Func &f);

        /// Add another function to this one
        Func &operator += (const Func &f);

        ~Func();

        /// Construct a new function on S_n that is identically zero
        static Func zero(int n);


        /// A string representation of this function
        std::string str();

        /// The actual mapping S_n -> R
        val_t operator () (const Permutation &sigma) const;

        /// Pointwise product of functions
        Func *operator * (const Func &other);

        /// Convolution
        Func *convolve(const Func &other);

        /// Compute a shifted function f so that f(p) = this(p0^-1 p p1)
        Func *shift(const Permutation &p0, const Permutation &p1);

        /// Set the val of this function at permutation p
        void set(const Permutation &p, val_t v);

        /// Compute the "naive" Fourier transform
        /*! This is a direct translation of the definition; it exists only
         *  for testing/demonstrating; in practice fft should be used instead
         *  \todo list running time
         */
        FourierFunc fourier() const;
        
        /// Compute the Fourier transform using Clausen's FFT
        /*! \todo list running time */
        FourierFunc fft() const;

        /// 
        Func restrict_to(int i) const;

        Func embed(int i) const;

        /// Return n where this a function over S_n
        int num_elements() const;

        /*! \brief Compute the probability that the tabloid (part, p0) maps to
         *    the tabloid (part, p1)
         */
        val_t marginalize(Partition part, Permutation p0, Permutation p1);

        /*! \brief Compute a matrix of marginals defined by the tabloids on
         *  partition part
         *
         *  The (i,j) entry of the matrix is the probability that tabloid i
         *   maps to tabloid j, where tabloids are enumerated in the order
         *   given by tabloid_list
         */
        Matrix marginals(Partition part);

        /*! \brief Compute the sum of f(p) over permutations p such that
         *  p(i) < p(j) (for a distribution, compute P(p(i) < p(j))
         */
        val_t ordered(int i, int j);

        /*! \brief Compute the matrix of order marginals:
         *  orderings()->(i,j) = ordered(i,j)
         */
        Matrix orderings();

        /*! \brief Normalize this distribution; public functions should always
         *  maintain normalization
         */
        void normalize();

        /*! \brief Construct the distribution formed by joining d0 and d1 as
         *  first-order independent
         *  \todo want arbitrary indices for d0, d1 */
        static Func *join(Func *d0, Func *d1);

        /*! \brief Split this distribution into two first-order independent
         *  distributions
         *
         * The first distribution will be a mapping over the indices is in the
         *  original distribution, and the second over {1, ..., n} \ is
         */
        std::pair<Func *, Func *> split(int n/*, int *is*/);
protected:
        /// Construct uninitialized func on S_n
        Func(int n);

        /// The explicit function: {0, ..., n! - 1} -> R
        /*! The indexing used is that defined by Permutation::index */
        val_t *func;

        /// This is a function over S_nelements
        int nelements;

        /// Number of values stored in func, i.e., nelements!
        int funclen;

        /// The maximum number of values displayed by the output of str
        static const int MAX_PRINT_ELEMENTS = 6;

        friend Func *uniform(int n);
        friend Func *mix_subset(int n, int k, Permutation::perm_val_t *subset);
        friend Func *mix_pair(int n, int i, int j, val_t p);
        friend Func *insertion_mix(int n);
        friend Func *mallows(Permutation p0, double c);
        friend Func *multitrack_obs(int n, int k, int *tracks, int *obs,
                                    Func::val_t p);
        friend Func *singletrack_obs(int n, int i, int j, Func::val_t pi);
        friend Func *bluetooth_obs(int n, int k, int *tracks, int *obs,
                           Func::val_t pi);
        friend Func *pair_rank_obs(int n, int i, int j, Func::val_t pi);

};




#endif
