/*! \file matrix.h
 *  \brief Declaration of an explicitly-stored matrix class
 */

#ifndef MATRIX_H
#define MATRIX_H

#include <exception>
#include <string>
#include <vector>
#include <itpp/itbase.h>

class DimensionMismatch : public std::exception {
public:
        const char *what() const throw ()
        {
                return "Dimension mismatch";
        }
};

/// A basic explicitly-stored matrix
class Matrix {
public:
        /// The type used for matrix entries
        typedef double mat_val_t;

        Matrix();
        Matrix(const Matrix &m);
        ~Matrix();

        Matrix &operator = (const Matrix &m);

        /// Construct an n x n identity matrix
        static Matrix identity(int n);

        /// Construct an r x c zero matrix
        static Matrix zero(int r, int c);

        Matrix operator * (mat_val_t s) const;

        Matrix operator += (const Matrix &m);
       
        mat_val_t ele(int i, int j) const;
        
        /// Compute the matrix dot product, trace(this^T*m)
        /*! Note that this is Theta(nm) while computing trace(this^T*m)
         *  explictly is O((nm)^2)
         */
        mat_val_t matrix_dot(const Matrix &m) throw (DimensionMismatch);

        /// Set the (r,c) element to val
        void set(int r, int c, mat_val_t val);

        Matrix submat(int r0, int c0, int r, int c) const;

        static Matrix direct_sum(std::vector<Matrix> ms);

        int num_rows() const;
        int num_cols() const;

        /// The trace of this matrix, i.e., the sum of the diagonal elements
        mat_val_t trace() const;

        std::string str() const;

/// \todo Fix encapsulation
//private:
        /// Construct an uninitialized r x c matrix
        Matrix(int r, int c);
        /// The matrix stored in row major order
        mat_val_t *mat;
        /// Number of rows in matrix
        int rows;
        /// Number of columns in matrix
        int cols;

};

itpp::mat direct_sum(std::vector<itpp::mat> blocks);


#endif
