/*! \file partition.h
 *  \brief Declarations of partitions and related things: tabloids and tableaux
 */

#ifndef PARTITION_H
#define PARTITION_H

#include "permutation.h"

#include <vector>

class InvalidPartition : public std::exception {
public:
        const char *what() const throw ()
        { return "Array is not a valid partition"; }
};

/// Partitions of integers represented explicitly
class Partition {
public:
        Partition();
        /// Create a new partition of sum(parts) consisting of p parts
        Partition(int p, int *parts) throw (InvalidPartition);

        /// Copy constructor
        Partition(const Partition &p);

        Partition &operator = (const Partition &p);

        ~Partition();

        /// Return the size of the ith greatest element of this partition
        int operator () (int i) const;

        /// The number that this partition partitions
        int num() const;

        /// The number of parts of this partition
        int num_parts() const;

        /// The hook length from the box at component i, position j
        int hook_len(int i, int j) const;

        /// \todo what should this be called?
        std::vector<Partition> minus() const;

        /// Form a new partition by decrementing the ith component of this one
        Partition decr(int i) const;

        Partition subpart(int start, int end) const;

        Partition concat(const Partition &part) const;

        /// Enumerate partitions in lexicographic order
        static std::vector<Partition> list(int n);

        static std::vector<Partition> list_up_to(int n, unsigned int size);

        /// Test two partitions for equality
        bool operator == (const Partition &p) const;

        /// Opposite of ==
        bool operator != (const Partition &p) const;

        /// Return true iff this partition dominates p
        /*! Domination is taken in the weak sense, so p.dominates(p)
         *  returns true
         */
        bool dominates(const Partition &p) const;

        /// A string representation of the partition
        std::string str() const;

        std::string repr() const;

private:
        /*! \brief Enumerate partitions in lexicographic order, where the first
         *  component is not to exceed max
         */
        static std::vector<Partition> list_up_to(int n, int max,
                                                 unsigned int size);

        /// The number being partitioned
        int n;
        /// The number of parts
        int p;
        /// The components of the partition
        int *parts;
}; 

/// (standard) Young tableau
class Tableau {
public:
        Tableau();
        /// Copy constructor
        Tableau(const Tableau &t);
        ~Tableau();

        /// Enumerate all standard tableau for a particular partition
        static std::vector<Tableau> list(const Partition &part);

        /// This tableau, written explicitly
        std::string str() const;

        // The (row, column) where the number i is
        std::pair<int, int> pos(int i) const;

        /*! \brief The axial distance, or (signed) number of steps required
         *  to get from i to j by moving in four adjacent directions
         */
        int axial_dist(int i, int j) const;

        /// Compute (i, i + 1) o (this tabloid)
        Tableau swap(int i) const;

        /// Test two tableau for equality
        bool operator == (const Tableau &t);

        /*! \brief Count the number of tableau of a particular shape using
         *  the hook formula
         */
        static int count(const Partition &part);
        
private:
        /// The column which contains the number i
        int col(int i) const;
        int *new_explicit_list() const;

        /// Construct an unitialized tableau on the given partition
        Tableau(const Partition &part);
        Partition part;
        int *tab;


};

/// Determine if two Young's tabloids are equal (as tabloids)
/*! The tabloids are both over the partition part, and defined to contain the
 *  elements p0(0), p0(1), ... filled in from left to right, largest component
 *  to smallest component
 */
bool tabloid_equal(const Partition &part,
                   const Permutation &p0,
                   const Permutation &p1);

/// List all the (tabloid_equal-distinct) Young's tabloids over a partition
/*! The tabloids are returned as permutations, as described in tabloid_equal
 */
std::vector<Permutation> tabloid_list(const Partition &part);

/// Calculate the multiplicity of an irrep in a given marginal rep
/*! This is the Kostka number, calculated according to Young's rule.
 *  m_part is the partition defining the marginal rep, and i_part is
 *  the partition defining the irrep
 */
int kostka(const Partition &m_part, const Partition &i_part);

#endif
