/*! \file permutation.cpp
 *  \brief Implementation of Permutation class
 */

#include <sstream>
#include <iostream>

#include "permutation.h"
#include "util.h"

using namespace std;

Permutation::Permutation()
 : map (0)
{

}

Permutation::Permutation(const Permutation &sigma)
{
        (*this) = sigma;
}

Permutation &Permutation::operator = (const Permutation &other)
{
	nelements = other.nelements;
	map = new perm_val_t[nelements];
	for (int i = 0; i < nelements; i++)
		map[i] = other.map[i];

        return *this;
}

Permutation Permutation::identity(int n)
{
	Permutation p;

	p.map = new perm_val_t[n];
	p.nelements = n;
	for (int i = 0; i < n; i++)
		p.map[i] = i;

	return p;
}

Permutation Permutation::swap(int n, int i, int j)
{
	/// \todo Check i, j
	Permutation p;

	p.map = new perm_val_t[n];
	p.nelements = n;
	for (int k = 0; k < n; k++) {
		if (k == i)
			p.map[k] = j;
		else if (k == j)
			p.map[k] = i;
		else
			p.map[k] = k;
	}

	return p;
}

/// \todo error checking
Permutation Permutation::from_inverse(int n, const int inv_map[])
        throw (InvalidPermutation)
{
        Permutation p;

        p.map = new int[n];
        p.nelements = n;

        for (int i = 0; i < n; i++)
                p.map[inv_map[i]] = i;

        return p;
}

int Permutation::min_cycle_ele() const
{
        int i;
        for (i = 0; i < nelements; i++)
                if (map[i] != i)
                        break;
        return i;
}

Permutation::Permutation(int n, const Permutation::perm_val_t _map[])
  throw (InvalidPermutation)
{
	if (!check_map(n, _map))
		throw InvalidPermutation();

	map = new perm_val_t[n];
	nelements = n;
	for (int i = 0; i < n; i++)
		map[i] = _map[i];
}

Permutation::~Permutation()
{
	delete [] map;
}

Permutation::perm_val_t Permutation::operator () (Permutation::perm_val_t i)
   throw (IndexOutOfBounds)
{
	if (i < 0 || i >= nelements)
		throw IndexOutOfBounds();
	return map[i];
}

Permutation Permutation::of(const Permutation &sigma) const
  throw (PermutationMismatch)
{
	if (nelements != sigma.nelements)
		throw PermutationMismatch();

	Permutation p;

	p.nelements = nelements;
	p.map = new perm_val_t[nelements];
	for (int i = 0; i < nelements; i++)
		p.map[i] = map[sigma.map[i]];

	return p;
}

Permutation Permutation::inverse() const
{
	Permutation p;

	p.nelements = nelements;
	p.map = new perm_val_t[nelements];
	for (int i = 0; i < nelements; i++)
		p.map[map[i]] = i;

	return p;
}

Permutation Permutation::direct_sum(const Permutation &sigma) const
{
        Permutation p;
        
        p.nelements = nelements + sigma.nelements;
        p.map = new perm_val_t[p.nelements];
        int i;
        for (i = 0; i < nelements; i++)
                p.map[i] = map[i];
        for (; i < nelements + sigma.nelements; i++)
                p.map[i] = sigma.map[i - nelements] + nelements;

        return p;
}

/// \todo check j >= i
Permutation Permutation::cont_cycle(int n, int i, int j)
{
        Permutation p;

        p.nelements = n;
        p.map = new perm_val_t[p.nelements];

        for (int k = 0; k < i; k++)
                p.map[k] = k;
        for (int k = i; k < j - 1; k++)
                p.map[k] = k + 1;
        if (i != j)
                p.map[j - 1] = i;
        for (int k = j; k < n; k++)
                p.map[k] = k;

        return p;
}

int Permutation::subindex(perm_val_t *left, int n) const
{
        perm_val_t last;
        perm_val_t *newleft;
        int lastind = -1;
        int offset;
        int ind;

        if (n == 0)
                return 0;

        last = map[n - 1];
        newleft = new perm_val_t[n - 1];
        offset = 0;
        for (int i = 0; i < n; i++) {
                if (left[i] == last) {
                        lastind = i;
                        offset = 1;
                } else {
                        newleft[i - offset] = left[i];
                }
        }

        ind = (n - 1 - lastind)*(int)fact(n - 1) + subindex(newleft, n - 1);

        delete [] newleft;

        return ind;
}

int Permutation::index() const
{
        return subindex(identity(nelements).map, nelements);
}

int Permutation::num_elements() const
{
        return nelements;
}

// this is not the obvious recursive algorithm, but instead works in place,
// without making any calls
Permutation Permutation::from_index(int n, int ind) throw (IndexOutOfBounds)
{
        Permutation p;
        p.nelements = n;
        p.map = new perm_val_t[n];

        // fill in the choices made at each step
        p.map[0] = 0;
        for (int i = 1; i < n; i++) {
                p.map[i] = i - ind % (i + 1);
                ind /= (i + 1);
        }

        // now find each number in the map
        for (int i = n - 1; i >= 1; i--) {
                int boost = 0;
                for (int j = n - 1; j >= 0; j--) {
                        if (p.map[j] + boost == i) {
                                p.map[j] = i;
                                break;
                        }
                        if (p.map[j] < i)
                                boost++;
                }
        }

        return p;
}

bool Permutation::setmapsto(int k, int *is, int *js)
{
	int m, n;
	for (m = 0; m < k; m++) {
		for (n = 0; n < k; n++) {
			if (js[m] == map[is[n]])
				break;
		}
		if (n == k) // no matches for this i
			break;
	}
	if (m == k) // match found for every i
		return true;
	return false;
}

string Permutation::str()
{
	ostringstream ret;
 
	ret << "[";
	for (int i = 0; i < nelements; i++)
		ret << " " << (*this)(i);
	ret << " ]";

	return ret.str();
}

string Permutation::repr()
{
	ostringstream ret;
 
	ret << "Permutation([";
	for (int i = 0; i < nelements; i++) {
                ret << (*this)(i);
                if (i != nelements - 1)
		        ret << ", ";
        }
	ret << "])";

	return ret.str();
}

vector<Permutation> Permutation::factor_swaps()
{
        Permutation p (*this);
        vector<Permutation> factors;

        for (int i = 0; i < nelements - 1; i++) {
                if (p(i) != i) {
                        factors.push_back(swap(nelements, p(i), i));
                        p = swap(nelements, p(i), i).of(p);
                }
        }

        return factors;
}

vector<Permutation> Permutation::factor_adj_swaps()
{
        Permutation p (*this);
        vector<Permutation> factors;

        for (int i = 0; i < nelements - 1; i++)
                for (int j = nelements - 1; j > i; j--) {
                        if (p.inverse()(j) < p.inverse()(j - 1)) {
                                factors.push_back(swap(nelements, j - 1, j));
                                p = swap(nelements, j - 1, j).of(p);
                        }
                }

        return factors;

}

bool Permutation::check_map(int n, const perm_val_t map[])
{
    if (n < 1)
            return false;

	// Note that this is Theta(n^2), but could be Theta(n log n)
	for (int i = 0; i < n; i++) {
		bool found = false;
		for (int j = 0; j < n; j++) {
			if (map[j] == i) {
				found = true;
				break;
			}
		}
		if (!found)
			return false;
	}

	return true;
}

int *new_seq(int i, int len)
{
        int *seq = new int[len];

        for(int j = 0; j < len; j++)
                seq[j] = j + i;

        return seq;
}

