/*! \file permutation.h
 *  \brief Definition of Permutation class
 */

#ifndef PERMUTATION_H
#define PERMUTATION_H

#include <string>
#include <vector>

class IndexOutOfBounds : public std::exception {
public:
        const char *what() const throw ()
        { return "Index out of bounds"; }
};

class InvalidPermutation : public std::exception {
public:
        const char *what() const throw ()
        { return "Array is not a valid permutation"; }
};

class PermutationMismatch : public std::exception {
public:
        const char *what() const throw ()
        { return "Permutations sizes do not match"; }
};

/// Permutations on n elements
/*! A permutation is a mapping from {0,...,n-1} to {0,...,n-1}, i.e.,
 *  an element of S_n.
 *  Permutations are immutable.
 */
class Permutation {
public:
        /// The type that this permutation maps
        typedef int perm_val_t;

        /// Construct a permutation given the explicit mapping
        Permutation(int n, const perm_val_t _map[]) throw (InvalidPermutation);

        /// Copy constructor
        Permutation(const Permutation &sigma);

        Permutation &operator = (const Permutation &other);

        ~Permutation();

        /// Construct the identity permutation on n elements
        static Permutation identity(int n);

        /// Construct the transposition of i and j
        static Permutation swap(int n, int i, int j);

        /*! \brief Construct a permutation from {0, ..., n-1} permuted,
            i.e., the inverse mapping
         */
        static Permutation from_inverse(int n, const int inv_map[])
                throw (InvalidPermutation);

        /// The minimum i that is not mapped to i
        /*! In other words, the first element of the first cycle
         *  comprising this permutation.
         *  In particular, if this permutation is (i, i+1), returns i.
         *  If this is the identity permutation, returns num_elements().
         */
        int min_cycle_ele() const;

        /// The result of applying this permutation to i
        perm_val_t operator () (perm_val_t i) throw (IndexOutOfBounds);

        /// Composition of permutations, i.e., this o sigma
        Permutation of(const Permutation &sigma) const
                throw (PermutationMismatch);

        /// The inverse permutation, this^-1
        Permutation inverse() const;

        /// The direct sum of permutations
        /*! If p0 is in S_n, p1 is in S_m, 
         *  the direct sum p0 (+) p1 = p is defined so that p(i) = p0(i) when
         *  i < n or p1(i - n) + n when i >= n
         */
        Permutation direct_sum(const Permutation &sigma) const;

        static Permutation cont_cycle(int n, int i, int j);

        /// Implements a bijection from S_n to {0, ..., n! - 1}
        /*! This is the inverse operation from from_index */
        int index() const;

        /// The number of elements permuted
        int num_elements() const;

        /// Implements a bijection from {0, ..., n! - 1} to S_n
        /*! This is the inverse operation from index */
        static Permutation from_index(int n, int i) throw (IndexOutOfBounds);

        /*! Decide if for every i in is there is some j in js such that
         *  p(i) = j, i.e, if the set of indices is maps unordered to the
         *  values js.
         *
         *  k should be the number of values pointed to by is and js.
         *  This function is useful for testing tabloid equivalence.
         */
        bool setmapsto(int k, int *is, int *js);

        /// Produce a "one line" explicit string representation
        std::string str();

        std::string repr();

        /// Factor this permutation into swaps (transpositions)
        std::vector<Permutation> factor_swaps();

        /*! \brief Factor this permutation into adjacent swaps
         *  (transpositions of the form (i,i+1))
         */
        std::vector<Permutation> factor_adj_swaps();

        /// For constructing by static methods
        Permutation();
private:

        /// Whether the mapping represents a valid element of S_n
        bool check_map(int n, const perm_val_t map[]);

        /// \todo document this
        int subindex(perm_val_t *left, int n) const;

        /// The mapping, stored explicitly: map[i] == this(i)
        perm_val_t *map;

        /// The number of elements mapped
        int nelements;

};

/// Return a newly allocated array of length len that counts from i
/*! The returned array should be freed with delete. */
int *new_seq(int i, int len);

#endif
